#include "MeshRenderer.h"
#include "Grid.h"
#include "UnstructuredGrid3D.h"
#include <GLUT/glut.h>									//GLUT library



void rainbow(float value,float& R,float& G,float& B)	//Maps a scalar in [0,1] to a color using the rainbow colormap
{
   const float dx=0.8f;

   value = (6-2*dx)*value+dx;
   R = max(0.0f,(3-(float)fabs(value-4)-(float)fabs(value-5))/2);
   G = max(0.0f,(4-(float)fabs(value-2)-(float)fabs(value-4))/2);
   B = max(0.0f,(3-(float)fabs(value-1)-(float)fabs(value-2))/2);
}

void grayscale(float value,float& R,float& G,float& B)	//Maps a scalar in [0,1] to a color using a grayscale colormap
{
   R = G = B = value;
}

void saturation(float value,float& R,float& G,float& B)	//Maps a scalar in [0,1] to a color using a saturation colormap
{
   float color_r = 0;									//The base color whose saturation we change (green).	
   float color_g = 1;									//Try different colors!	
   float color_b = 0;
   
   if (value<0.5)										//value in [0,0.5]: modulate the luminance from black to the base-color.
   {   
	   R = 2*value*color_r;
	   G = 2*value*color_g;
	   B = 2*value*color_b;
   }
   else													//value in [0.5,1]: modulate the saturation from base-color to white.
   {
	   value = 2*(value-0.5);	
	   R = (1-value)*color_r + 1*value;
	   G = (1-value)*color_g + 1*value;
	   B = (1-value)*color_b + 1*value;
   }
}





void MeshRenderer::draw(Grid& g_)						//Draw an unstructured grid, in the style indicated by 'drawing_style'
{
	glClearColor(1,1,1,1);								//1. Clear the frame and depth buffers
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_NORMALIZE);								//2. Ask OpenGL to automatically normalize normal-vectors when we scale the model
														//   This is required to get correct shading for various model scales.

	UnstructuredGrid3D& g = (UnstructuredGrid3D&)g_;
	
	switch (draw_style)									//3. Render the grid, based on the current drawing style
	{
	case DRAW_GRID: 
		glColor3f(0.5,0.5,0.5);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glEnable(GL_POLYGON_OFFSET_FILL);				//   Use OpenGL polygon offsetting to draw the filled polygon mesh a bit 	
		glPolygonOffset(1.0,1.0);						//   behind the wireframe drawing. This makes the two overlay nicely.
		drawGrid(g);
		glDisable(GL_POLYGON_OFFSET_FILL);
		glColor3f(0,0,0);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		drawGrid(g);
		break;
	case DRAW_POINTS:
		drawPoints(g);
		break;
	case DRAW_C1_CELLS:
		drawC1Cells(g);
		break;
	}			
	
	glutSwapBuffers();
}



void MeshRenderer::drawPoints(UnstructuredGrid3D& g)	//Draw 'g' as a point cloud
{
	glDisable(GL_LIGHTING);								//     Disable lighting, since we cannot use two-side lighting for point clouds
	glPointSize(2);										//1.   Draw all vertices in the grid as large points

	ScalarAttributes& scalars = g.pointScalars();		//     Get scalar range, to normalize scalar-values next in [0,1]
	float scalar_min = scalars.getMin();
	float scalar_max = scalars.getMax();	
	float scalar_range = scalar_max-scalar_min;
	if (scalar_range==0) scalar_range=1;

	glBegin(GL_POINTS);									//2.   Draw vertices as an OpenGL point-set	
	for(int i=0;i<g.numPoints();++i)						
	{
		float v[3];
		g.getPoint(i,v);								//3. Get coordinates and normal of i-th vertex			
		float  scalar = (scalars.getC0Scalar(i)-scalar_min)/scalar_range;
		float R,G,B;
		rainbow(scalar,R,G,B);

		glColor3f(R,G,B);															
		glVertex3f(v[0],v[1],v[2]);						//2.2.  Pass color, then coordinate, to OpenGL
	}
	glEnd();											//3.    Finish rendering the entire point-set
}




void MeshRenderer::drawGrid(UnstructuredGrid3D& g)		//Draw 'g' without shading
{	
	glDisable(GL_LIGHTING);

	for(int i=0;i<g.numCells();++i)						//1. Draw all cells in the grid
	{
		int   cell[10];									//
		
		int nvertices = g.getCell(i,cell);				//2.   Get the 'nvertices' vertex-IDs of i-th cell
		if (nvertices!=3 && nvertices!=4)				//     We only handle here drawing of triangle and quad cells.
		   continue;									//     It is quite simple to extend this to other cell types.

		if (nvertices==3)								//     Triangle cells:
			glBegin(GL_TRIANGLES);						
		else //nvertices==4	
			glBegin(GL_QUADS);							//     Quad cells:		

		for(int j=0;j<nvertices;++j)					//3.   Render current cell
		{
			float x[3];
			g.getPoint(cell[j],x);						//     Get vertex coordinates of j-th vertex of i-th cell		
			glVertex3f(x[0],x[1],x[2]);					//     Pass this coordinate to OpenGL
		}	
		glEnd();                                        //4.   Finish rendering current cell
	}


	glDisable(GL_LIGHTING);								//5.   Disable lighting, since we cannot use two-side lighting for point clouds
	glPointSize(5);										//     Draw all vertices in the grid as large points
	glEnable(GL_POLYGON_OFFSET_POINT);					//     Use OpenGL point offsetting to draw the selected points a bit in front of 	
	glPolygonOffset(1.0,-1.0);							//     the mesh drawing. This makes the two overlay nicely.


	ScalarAttributes& scalars = g.pointScalars();		//     Get scalar range, to normalize scalar-values next in [0,1]
	float scalar_min = scalars.getMin();
	float scalar_max = scalars.getMax();	
	float scalar_range = scalar_max-scalar_min;
	if (scalar_range==0) scalar_range=1;


	glBegin(GL_POINTS);									//6.   Draw selected vertices (having a non-zero scalar value) as an OpenGL point-set	
	for(int i=0;i<g.numPoints();++i)						
	{
		float v[3];
		g.getPoint(i,v);								//7.   Get coordinates and normal of i-th vertex			
		float  scalar = scalars.getC0Scalar(i);
		if (!scalar) continue;
		scalar = (scalar-scalar_min)/scalar_range;
		float R,G,B;
		rainbow(scalar,R,G,B);

		glColor3f(R,G,B);
		glVertex3f(v[0],v[1],v[2]);						//8.   Pass color, then coordinate, to OpenGL
	}
	glEnd();											//9.   Finish rendering the entire point-set

	glEnable(GL_POLYGON_OFFSET_POINT);
}




void MeshRenderer::drawC1Cells(UnstructuredGrid3D& g)	//Draw 'g' with smooth shading (bilinear interpolation of shading at vertices)
{	
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);			//3. Render cells filled
	glEnable(GL_COLOR_MATERIAL);
  	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
	glShadeModel(GL_SMOOTH);
	
	
	ScalarAttributes& scalars = g.pointScalars();		//     Get scalar range, to normalize scalar-values next in [0,1]
	float scalar_min = scalars.getMin();
	float scalar_max = scalars.getMax();	
	float scalar_range = scalar_max-scalar_min;
	if (scalar_range==0) scalar_range=1;
	

	VectorAttributes& point_normals = g.getPointNormals();

	for(int i=0;i<g.numCells();++i)						//1. Draw all cells in the grid 
	{
		int   cell[10];									//
		
		int nvertices = g.getCell(i,cell);				//2.   Get the 'nvertices' vertex-IDs of i-th cell
		if (nvertices!=3 && nvertices!=4)				//     We only handle here drawing of triangle and quad cells.
		   continue;									//     It is quite simple to extend this to other cell types.

		if (nvertices==3)								//     Triangle cells:
			glBegin(GL_TRIANGLES);						
		else //nvertices==4	
			glBegin(GL_QUADS);							//     Quad cells:
		

		for(int j=0;j<nvertices;++j)					//3.   Render current cell
		{
			float x[3];									//     Get vertex coordinates of j-th vertex of i-th cell
			g.getPoint(cell[j],x);						//     Next, get vertex normal of the same vertex			
			float* point_normal = point_normals.getC0Vector(cell[j]);	
			float  scalar = (scalars.getC0Scalar(cell[j])-scalar_min)/scalar_range;
			float R,G,B;
			rainbow(scalar,R,G,B);

			glColor3f(R,G,B);		
			glNormal3f(point_normal[0],point_normal[1],point_normal[2]);
			glVertex3f(x[0],x[1],x[2]);					//     Pass the vertex normal, then the vertex coordinate, to OpenGL
		}	
		glEnd();                                        //4.   Finish rendering current cell
	}
}




