#include <GLUT/glut.h>									//GLUT library
#include <iostream>
#include "UnstructuredGrid3D.h"							//the 3D unstructured grid class
#include "MeshRenderer.h"								//a simple renderer for 3D unstructured grids, demonstrating flat/smooth shading
#include "PlyReader.h"									//a reader that initializes UnstructuredGrid3D objects with meshes stored in the PLY file format
#include "zpr.h"										//library for interactively manipulating the OpenGL camera (viewpoint) using the mouse
#include "PointSearch.h"


float fov	 = 60;										//Perspective projection parameters
float z_near = -100; 
float z_far  = 100;


UnstructuredGrid3D*	grid = 0;							//The 3D unstructured grid (mesh) we work on.
MeshRenderer		renderer;							//Grid renderer engine. Renders grid with selected points using various colormaps
PointSearch*		point_search = 0;					//Engine to search closest points to a given 3D point
int					drawing_style = 0;					//Drawing style for grid and closest points.
int					selection_mode = 0;					//Selection mode:
														//  0: draw selected points in one color, and unselected points in another color
														//	1: draw grid colored by distance of selected points to search point
Point3d*			picked_point = 0;					//Point selected on grid by interactive picking



void viewing(int W, int H)								//Window resize function, sets up viewing parameters (GLUT callback function)
{
	glMatrixMode(GL_PROJECTION);						//1. Set the projection matrix
	glLoadIdentity();	
	gluPerspective(fov,float(W)/H,z_near,z_far);

	glViewport(0,0,W,H);								//2. Set the viewport to the entire size of the rendering window
}



void mouseclick(int button, int state, int x, int y)	//Callback for mouse click events. We use these to control the viewpoint interactively.
{
  int keys = glutGetModifiers();						//The mouse events are interpreted as follows:	
  if (keys & GLUT_ACTIVE_CTRL)							// 
    button = GLUT_MIDDLE_BUTTON;						// -left button + move:                         rotate viewpoint
  if (keys & GLUT_ACTIVE_SHIFT)							// -middle button (or left button+Control key): zoom in / out
    button = GLUT_RIGHT_BUTTON;							// -right buttom (or left button+Shift key):    translate viewpoint
  
  zprMouse(button,state,x,y);							//Use the ZPR library to manipulate the viewpoint. The library sets the modelview  
														//OpenGL matrix based on the mouse events, thereby changing the current viewpoint.
}



void mousemotion(int x, int y)							//Callback for mouse move events. We use these to control the viewpoint interactively.
{
  zprMotion(x,y);										//Pass the current location of the mouse to the ZPR library, to change the viewpoint.

  glutPostRedisplay();									//After each mouse move event, ask GLUT to redraw our window, so we see the viewpoint change.
}



void findClosestPoints(const Point3d& p)				//
{
	vector<int>   indexes;
	vector<float> distances;
	const int Nclosest = 300;
	
	point_search->closest(Nclosest,p,indexes,distances);

	ScalarAttributes& scalars = grid->pointScalars();
	scalars.zero();
	
	for(int i=0;i<Nclosest;++i)
	{
	   float value = (selection_mode==0)? 1 : distances[i];	
	   scalars.setC0Scalar(indexes[i],value);
	}
}



void pick(double x,double y,double z)					//Callback for mouse click (picking) events. Finds closest points to picked point
{														//and sets the grid vertex scalar values to show these (or distances to picked point).
	delete picked_point;								//Delete earlier picked point (if any), since we have a new onw
	
	picked_point = new Point3d(x,y,z);					//Remember which was the picked point

	findClosestPoints(*picked_point);					//Find closest points to picked point and set the grid vertex-scalars to reflect identities or distances to picked point
}




void keyboard(unsigned char c,int,int)					//Callback for keyboard events:
{
	switch(c)	
	{
	case ' ':											// space:   Toggle through the various drawing styles of the mesh renderer
		{
		  drawing_style = (++drawing_style) % 3;		
		  renderer.setDrawingStyle((MeshRenderer::DRAW_STYLE)drawing_style);
		  break;
		}
	case 'R':											// 'r','R': Reset the viewpoint
	case 'r':
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		zprInit(0,0,0);
		break;
	case 'C':
	case 'c':
		selection_mode = (++selection_mode) % 2;	
		if (picked_point) findClosestPoints(*picked_point);
		break;
	}
	
	glutPostRedisplay();			
}	

void draw()												//Render the 3D mesh (GLUT callback function)
{
	renderer.draw(*grid);
}




int main(int argc,char* argv[])							//Main program
{
   cout<<"3D unstructured grid (mesh) visualization."<<endl;
   cout<<"Interaction options: "<<endl<<endl;
   cout<<"  Mouse: changes the viewpoint:"<<endl;
   cout<<"      -rotate:     left-click and drag"<<endl;
   cout<<"      -translate:  right-click (or Shift-left-click) and drag"<<endl;
   cout<<"      -scale:      middle-click (or Control-left-click) and drag"<<endl;
   cout<<"      -click:      select N closest grid points"<<endl; 
   cout<<"  Keyboard: visualization options:"<<endl;
   cout<<"      -r,R:        reset the viewpoint"<<endl;
   cout<<"      -space:      cycle through mesh rendering styles"<<endl;
   cout<<"      -c,C:        color selected points by distance to seed point or selection status"<<endl;

   const char* filename = (argc<2)? "DATA/scapula.ply" : argv[1];  //Read the PLY file given as 1st argument. If no arguments given, use a default file.

   glutInit(&argc, argv);								//1.  Initialize the GLUT toolkit
   glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
														//2.  Ask GLUT to create next windows with a RGB framebuffer and a Z-buffer too
   glutInitWindowSize(500,500);							//3.  Tell GLUT how large are the windows we want to create next
   glutCreateWindow("6. 3D mesh (unstructured grid)");	//4.  Create our window
   zprInit(0,0,0);										//5.  Initialize the viewpoint interaction-tool to look at the point (0,0,0)
   zprPickFunc(pick);
   
   PlyReader rdr;										//6.  Read a 3D mesh stored in a file in the PLY format
   grid = rdr.read(filename);
   grid->normalize();									//7.  Normalize the mesh in the [-1,1] cube. This makes setting the OpenGL projection easier.
   grid->computeFaceNormals();							//8.  Compute face and vertex normals for the mesh. This allows us to shade the mesh next.
   grid->computeVertexNormals();
   point_search = new PointSearch(*grid,1000);			//9.  Create a point-search engine to search closest grid vertices to any 3D point. 
														//    We want to search up to 1000 closest-points.
	  	  
   glutMouseFunc(mouseclick);							//10. Bind the mouse click and mouse drag (click-and-move) events to callbacks. This allows us
   glutMotionFunc(mousemotion);							//    next to control the viewpoint interactively.
   glutKeyboardFunc(keyboard);
   glutDisplayFunc(draw);								//11. Add a drawing callback to the window	
   glutReshapeFunc(viewing);							//12. Add a resize callback to the window
   glutMainLoop();										//13. Start the event loop that displays the graph and handles window-resize events
   
   return 0;
}




